// Copyright 2015 The Shaderc Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "file.h"

namespace glslc {

shaderc_util::string_piece GetFileExtension(
    const shaderc_util::string_piece& filename) {
  size_t dot_pos = filename.find_last_of(".");
  if (dot_pos == shaderc_util::string_piece::npos) return "";
  return filename.substr(dot_pos + 1);
}

}  // namespace glslc
